#include <string.h>

#include <vhttpd/http_message.h>
#include <vhttpd/channel.h>
#include <vhttpd/http_connection.h>

#include "private.h"

//debug
#include <stdio.h>

/*** types */

struct connection_t
{
  int active_;
  channel_t *channel_;
  connection_handler_t handler_;
  message_t *request_;		/* current prepared request */
  buffer_t *buffer_;
  void *data_;
};

/*** helpers */

static void
create_request (connection_t * conn)
{
  if (conn && !conn->request_)
    conn->request_ = create_request_message (NULL,NULL);
}

static void
destroy_request (connection_t * conn)
{
  if (!conn)
    return;

  destroy_message (conn->request_);
  conn->request_ = NULL;
}

static int
is_keep_alive (message_t * res, message_t * req)
{
  const char *c;

  return res &&
    (message_get_code (res) < bad_request) &&
    (c = message_get_header (req, "connection")) && !strcmp (c, "keep-alive");
}

static void
connection_close (connection_t * conn)
{
  conn->active_ = FALSE;
}

/* remove port from host header */
static void
cleanup_host (message_t * msg)
{
  const char *host = message_get_header (msg, "host");

  if (host)
    {
      const char *pos = strchr (host, ':');
      char *host2 =
	g_strndup (host, pos ? (size_t) (pos - host) : strlen (host));
      message_set_header (msg, "host", host2);
      g_free (host2);
    }
}

/*** functions */
connection_t *
connection_new (channel_t * channel, connection_handler_t handler, void *data)
{
  connection_t *conn;

  if (!channel)
    return NULL;

  conn = g_new0 (connection_t, 1);

  conn->channel_ = channel;
  conn->handler_ = handler;
  conn->buffer_ = g_string_new ("");
  conn->active_ = TRUE;
  conn->data_ = data;

  return conn;
}


void
connection_delete (connection_t * conn)
{
  if (!conn)
    return;

  g_string_free (conn->buffer_, TRUE);
  destroy_message (conn->request_);
  g_free (conn);
}

channel_t *
connection_get_channel (connection_t * conn)
{
  return conn->channel_;
}

int
connection_send_message (connection_t * conn, const message_t * msg)
{
  int retcode = 0;
  if (conn && conn->active_ && msg)
  {
    channel_t *c = conn->channel_;
    GString *startline = message_write_startline(msg);
    GString *header;
    const char *body = message_get_body (msg);
    int fd = message_get_file (msg);
    int len = message_get_length (msg);

    /* host address is required for http 1.1 */
    header = message_write_header(msg,channel_get_remoteaddr(c));

    if (channel_send (c, startline->str, startline->len) ||
        channel_send (c, header->str, header->len) ||
        (body && channel_send (c, body, len)) ||
        ((fd >= 0) && channel_sendfile (c, fd, len)))
        {
          retcode = 1;
	  connection_close (conn); /* close on errors during message sending */
	}

    g_string_free (startline, TRUE);
    g_string_free (header, TRUE);
  }
  return retcode;
}

retcode_t
connection_receive_message (connection_t * conn, message_t * msg)
{
  retcode_t retcode;

  if (!conn || !conn->active_ || !msg)
    return bad_request;

  /* receive next portion of data */
  if (channel_receive (conn->channel_, conn->buffer_))
    connection_close (conn); /* end of data */

  /* parse current portion of data */
  if ((retcode = message_read(msg, conn->buffer_)) != unfinished)
  {
    /* transfer some useful information from channel to request */
    message_set_plain (msg, channel_get_type (conn->channel_) == CHANNEL_TYPE_PLAIN);
    message_set_header (msg, "remote-addr", channel_get_remoteaddr (conn->channel_));
    /* try to fix "host" header value, remove port if necessary */
    cleanup_host (msg);
  }
  return retcode;
}


void
connection_read (connection_t * conn)
{
  retcode_t retcode = ok;

  if (!conn || !conn->active_)
    return;

  create_request (conn);

  if ((retcode = connection_receive_message(conn,conn->request_)) != unfinished)
    {
      message_t *response = NULL;
      int keep_alive;

      if (conn->handler_)
	response = conn->handler_ (retcode, conn->request_, conn->data_);
      if (!response)
	response = create_response_message (retcode);

      keep_alive = is_keep_alive (response, conn->request_);
      message_set_header (response, "connection",
			  (keep_alive) ? "keep-alive" : "close");
      connection_send_message (conn, response);

      if (!keep_alive)
	connection_close (conn);

      destroy_message (response);
      destroy_request (conn);
    }
}

int
connection_is_open (connection_t * conn)
{
  return conn ? conn->active_ : FALSE;
}
