#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>

#include <string.h>

#include <glib.h>

#include <vhttpd/time.h>
#include <vhttpd/http_message.h>

#include "private.h"

/*** constants */

const char _LWS[] = " \t";
const char _CRLF[] = "\r\n";

/*** helpers */

static void
_clear_content (message_t * r)
{
  if (r->body_)
    {
      g_free (r->body_);
      r->body_ = NULL;
    }

  if (r->fd_ >= 0)
    {
      close (r->fd_);
      r->fd_ = -1;
    }
  r->length_ = 0;
}

static void
_set_header (message_t * r, const char *name, char *value)
{
  if (!r || !name || !name[0] || !value || !value[0])
    return;

  g_hash_table_replace (r->headers_,
			g_ascii_strdown (name, (gssize) strlen (name)),
			value);
}

static void
_set_cookie(message_t * r, const char *name, char *value)
{
  if (!r || !name || !name[0] || !value || !value[0])
    return;

  g_hash_table_replace (r->cookies_, g_strdup (name), value);
}

static void
_set_string(char **obj, const char *value)
{
    g_free(*obj);
    *obj = value?g_strdup(value):NULL;
}

/*** functions */

message_t *
create_message (void)
{
  message_t *r = g_new0 (message_t, 1);

  r->headers_ =
    g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
  r->cookies_ =
    g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
  r->fd_ = -1;
  r->length_ = 0;
  r->state_ = parse_request_line;

  message_set_header (r, "content-length", "0");

  return r;
}

message_t *
create_request_message(const char *method, const char *uri)
{
  message_t *r = create_message ();

  r->type_ = message_request;
  message_set_method(r,method);
  message_set_uri(r,uri);

  return r;
}

message_t *
create_response_message (retcode_t code)
{
  message_t *r = create_message ();
  char *cur_time = time_string (time (NULL));

  r->type_ = message_response;
  r->code_ = code;

  message_set_header (r, "date", cur_time);
  message_set_header (r, "last-modified", cur_time);
  message_set_header (r, "content-type", "text/html;charset=utf8");
  message_set_header (r, "cache-control", "no-cache");

  g_free (cur_time);

  return r;
}

void
destroy_message (message_t * r)
{
  if (!r)
    return;

  _clear_content(r);

  /* common part */
  g_free (r->version_);
  g_hash_table_destroy (r->headers_);
  g_hash_table_destroy (r->cookies_);

  /* request part */
  g_free (r->method_);
  g_free (r->uri_);
  g_free (r->query_);
  g_free (r->startline_);

  /*request helpers */
  g_free (r->lastheader_);

  g_free (r);
}



void
message_set_header_vprintf (message_t * r, const char *name, const char *fmt,
		     va_list args)
{
  if (!r || !name || !name[0] || !fmt || !fmt[0])
    return;

  _set_header(r,name, g_strdup_vprintf (fmt, args));
}

void
message_set_header_printf (message_t * r, const char *name, const char *fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  message_set_header_vprintf (r, name, fmt, ap);
  va_end (ap);
}

void
message_set_header (message_t * r, const char *name, const char *value)
{
  if (!r || !name || !name[0] || !value || !value[0])
    return;

  _set_header(r,name,g_strdup(value));
}

const char *
message_get_header (const message_t * r, const char *name)
{
  char *key;
  char *result;

  if (!r || !name || !name[0])
    return NULL;

  key = g_ascii_strdown (name, (gssize) strlen (name));
  result = g_hash_table_lookup (r->headers_, key);

  g_free (key);
  return result;
}

void
message_set_cookie_vprintf (message_t * r, const char *name, const char *fmt, va_list args)
{
  if (!r || !name || !name[0] || !fmt || !fmt[0])
    return;

  _set_cookie(r,name,g_strdup_vprintf (fmt, args));
}

void
message_set_cookie_printf (message_t * r, const char *name, const char *fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  message_set_cookie_vprintf (r, name, fmt, ap);
  va_end (ap);
}

void
message_set_cookie (message_t * r, const char *name, const char *value)
{
  if (!r || !name || !name[0] || !value || !value[0])
    return;

  _set_cookie(r,name,g_strdup(value));
}


const char *
message_get_cookie (const message_t * r, const char *name)
{
  if (!r || !name || !name[0])
    return NULL;

  return g_hash_table_lookup (r->cookies_, name);
}

char *
message_header_field (char *line, char **saveptr)
{
  char *result = strtok_r (line, _LWS, saveptr);

  return result ? g_strdup (result) : NULL;
}

const char *
message_get_body (const message_t * r)
{
  return r ? r->body_ : NULL;
}

int
message_get_file (const message_t * r)
{
  return r ? r->fd_ : -1;
}

int
message_get_length (const message_t * r)
{
  return r ? r->length_ : 0;
}

void
message_set_body (message_t * r, const char *data, size_t length)
{
  if (!r || !data)
    return;

  _clear_content (r);

  r->length_ = length;
  if (length > 0)
  {
    r->body_ = g_new0 (char, length);
    memcpy (r->body_, data, length * sizeof(char));
  }

  message_set_header_printf (r, "content-length", "%zu", length);
}

void
message_set_file (message_t * r, int fd, size_t length)
{
  if (!r || (fd < 0))
    return;

  _clear_content (r);

  r->length_ = length;
  r->fd_ = dup (fd);
  close_on_exec(r->fd_);

  message_set_header_printf (r, "content-length", "%zu", length);
}

const char *
message_get_startline (const message_t * r)
{
  return r ? r->startline_ : "";
}

void
message_set_startline (message_t * r, const char *data)
{
  if (r) _set_string(&r->startline_,data);
}

int
message_get_plain (const message_t * m)
{
  return m ? m->plain_ : TRUE;
}

void
message_set_plain (message_t * m, int plain)
{
  if (m)
    m->plain_ = plain;
}

const char *
message_get_uri (const message_t * r)
{
  return r ? r->uri_ : "";
}

void
message_set_uri (message_t * r, const char *value)
{
    if (r) _set_string(&r->uri_,value);
}

const char *
message_get_query (const message_t * r)
{
  return r ? r->query_ : "";
}

void
message_set_query (message_t * r, const char *value)
{
    if (r) _set_string(&r->query_,value);
}

const char *
message_get_method (const message_t * r)
{
  return r ? r->method_ : "";
}

void
message_set_method (message_t * r, const char *value)
{
    if (r) _set_string(&r->method_,value);
}

retcode_t
message_get_code (const message_t * r)
{
  return r ? r->code_ : bad_request;
}

