#define _GNU_SOURCE
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <sys/select.h>

#include <errno.h>

#include <glib.h>

#include <vhttpd/log.h>
#include <vhttpd/channel.h>
#include <vhttpd/tls_channel.h>
#include <vhttpd/smart_channel.h>
#include <vhttpd/http_server.h>

//debug
#include <stdio.h>

/*** types */

struct server_t
{
  channel_t *ch_;
  GSList *cons_;
  void *data_;

  /* select helpers */
  int nfds_;
  fd_set rfds_;
};

/*** helpers */

static connection_t *
connection_open (channel_t * ch, connection_handler_t handler, void *data)
{
  return connection_new (channel_accept (ch), handler, data);
}

static void
connection_close (gpointer conn, gpointer data __attribute__ ((__unused__)))
{
  destroy_channel (connection_get_channel ((connection_t *) conn));
  connection_delete ((connection_t *) conn);
}

static int
connection_get_fd (connection_t * conn)
{
  return channel_get_fd (connection_get_channel (conn));
}

static void
set_fd (gpointer conn, gpointer data)
{
  int fd = connection_get_fd ((connection_t *) conn);
  server_t *serv = (server_t *) data;

  FD_SET (fd, &(serv->rfds_));
  serv->nfds_ = MAX (fd, serv->nfds_);
}

static void
check_fd (gpointer conn, gpointer data)
{
  int fd = connection_get_fd ((connection_t *) conn);
  server_t *serv = (server_t *) data;

  if (FD_ISSET (fd, &(serv->rfds_)))
    {
      connection_read ((connection_t *) conn);
    }
}


static GSList *
c_list_append (GSList * s, connection_t * c)
{
  return c ? g_slist_prepend (s, c) : s;
}

static GSList *
c_list_filter (GSList * s)
{
  GSList *in = s;
  GSList *out = NULL;

  for (; in; in = g_slist_next (in))
    {
      if (connection_is_open ((connection_t *) in->data))
	out = g_slist_prepend (out, in->data);
      else
	connection_close ((connection_t *) in->data, NULL);
    }
  g_slist_free (s);

  return out;
}


static void
c_list_free (GSList * s)
{
  g_slist_foreach (s, connection_close, NULL);
  g_slist_free (s);
}

static server_t *
create_server (channel_t * ch)
{
  server_t *serv;

  if (!ch)
    return NULL;

  serv = g_new0 (server_t, 1);
  serv->ch_ = ch;

  return serv;
}

/*** functions */

void
destroy_server (server_t * serv)
{
  if (!serv)
    return;

  destroy_channel(serv->ch_);
  c_list_free (serv->cons_);
  g_free (serv);
}

void
server_loop (server_t * serv, connection_handler_t handler, void *data)
{
  int ch;
  fd_set *rfds;

  if (!serv || ((ch = channel_get_fd (serv->ch_)) == -1))
    return;

  rfds = &(serv->rfds_);

  while (TRUE)
    {
      struct timeval tv;
      int retval;

      FD_ZERO (rfds);
      FD_SET (ch, rfds);

      serv->nfds_ = ch;
      g_slist_foreach (serv->cons_, set_fd, serv);

      tv.tv_sec = 10;
      tv.tv_usec = 0;

      retval =
	TEMP_FAILURE_RETRY (select (serv->nfds_ + 1, rfds, NULL, NULL, &tv));

      if (retval == -1)
	RUNTIME_ERROR ("select:%m");
      else if (retval == 0)
	continue;

      if (FD_ISSET (ch, rfds))
	serv->cons_ = c_list_append (serv->cons_,
				      connection_open (serv->ch_, handler,
							data));

      g_slist_foreach (serv->cons_, check_fd, serv);
      serv->cons_ = c_list_filter (serv->cons_);
    }
}

server_t *
create_server_from_socket (int socket_fd)
{
  return create_server(create_channel_from_socket(socket_fd, NULL, C_SO_DEFAULT_SERVER));
}

server_t *
create_tls_server_from_socket (int socket_fd, const char *keyfile, const char *certfile)
{
  return create_server(create_tls_channel_from_socket(socket_fd, NULL, keyfile, certfile, C_SO_DEFAULT_SERVER));
}

server_t *
create_smart_server_from_socket (int socket_fd, const char *keyfile, const char *certfile)
{
  return create_server(create_smart_channel_from_socket(socket_fd, NULL, keyfile, certfile, C_SO_DEFAULT_SERVER));
}

server_t *
create_inet_server (const char *addr, const char* port)
{
  return create_server(create_inet_channel(addr,port,C_SO_DEFAULT_SERVER));
}

server_t *
create_unix_server (const char *path)
{
  server_t * srv=create_server(create_unix_channel(path,C_SO_DEFAULT_SERVER));
  return srv;
}

server_t *
create_tls_server (const char *addr, const char* port, const char *keyfile, const char *certfile)
{
  return create_server(create_tls_channel(addr, port, keyfile, certfile, C_SO_DEFAULT_SERVER));
}

server_t *
create_smart_server (const char *addr, const char* port, const char *keyfile, const char *certfile)
{
  return create_server(create_smart_channel(addr, port, keyfile, certfile, C_SO_DEFAULT_SERVER));
}
