#define _GNU_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include <netdb.h>
#include <sys/sendfile.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/stat.h>

#include <sys/un.h>
#include <netinet/in.h>

#include <pwd.h>
#include <grp.h>

#include <errno.h>

#include <glib.h>

#include <vhttpd/log.h>
#include <vhttpd/socket.h>

#include "private.h"

#define NLISTEN 100

/*** helpers */

#define EXIT_IF_FATAL(flags, message) \
    if (FLAG_ISSET((flags), C_SO_FATAL)) \
        RUNTIME_ERROR (#message)

typedef int (*bind_or_connect_func)(int, const struct sockaddr *,socklen_t);

static int
set_nonblock (int fd)
{
  int value = 1;

  return ioctl (fd, (unsigned long int) FIONBIO, &value);
}

static const char *
any (const char *addr)
{
  return (!addr || !*addr || *addr == '*') ? NULL : addr;
}

static void
fill_sockaddr_un(struct sockaddr_un *addr,const char *path)
{
    memset(addr, 0, sizeof(struct sockaddr_un));
    addr->sun_family = AF_UNIX;

    memset(addr->sun_path,0,sizeof(addr->sun_path));
    strncpy(addr->sun_path,path,sizeof(addr->sun_path)-1);
}

static int
connect_unix(int s, const char *path)
{
    struct sockaddr_un addr;
    fill_sockaddr_un(&addr,path);
    return connect(s, (struct sockaddr *)&addr, SUN_LEN(&addr));
}

static int
bind_unix(int s,const char *path)
{
    struct sockaddr_un addr;

    fill_sockaddr_un(&addr,path);
    if ((bind(s, (struct sockaddr *)&addr, SUN_LEN(&addr)) < 0) && (errno == EADDRINUSE))
    {
	if (connect_unix(s,path) == 0)
	    RUNTIME_ERROR ("address %s in use",addr.sun_path);
	/* unlink path from previous bind */
	if (unlink(path) < 0)
	    RUNTIME_ERROR ("unlink:%m");
	/* bind again */
	if (bind(s, (struct sockaddr *)&addr, SUN_LEN(&addr)) < 0)
	    RUNTIME_ERROR ("bind:%m");
    }
    return 0;
}

static int
_create_inet_socket (const char *node_addr, const char *service, int flags, const struct addrinfo *hints)
{
  struct addrinfo *ai, *rp;
  int retcode;
  int fd = -1;
  int value = 1;
  bind_or_connect_func bind_or_connect = connect;

  if ((retcode = getaddrinfo (node_addr, service, hints, &ai)) != 0)
    {
      if (retcode == EAI_ADDRFAMILY ||
          retcode == EAI_FAMILY ||
          retcode == EAI_NONAME ||
          retcode == EAI_NODATA ||
          retcode == EAI_AGAIN)
        {
          /* Non-fatal error */
          return -1;
        }
        RUNTIME_ERROR ("getaddrinfo:%s", gai_strerror (retcode));
    }

  for (rp = ai; rp != NULL; rp = rp->ai_next)
    {
      fd = socket (rp->ai_family, rp->ai_socktype, rp->ai_protocol);

      if (fd == -1)
          continue;

      close_on_exec(fd);

      if (FLAG_ISSET(flags, C_SO_SERVER))
      {
          bind_or_connect = bind;
          if (setsockopt (fd, SOL_SOCKET, SO_REUSEADDR,
                          &value, (socklen_t) sizeof (value)))
              RUNTIME_ERROR ("setsockopt:%m");
      }

      if (bind_or_connect(fd, rp->ai_addr, rp->ai_addrlen) == 0)
          break;

      close (fd);
      fd = -1;
    }

  freeaddrinfo (ai);

  return fd;
}

/*** functions */
int
create_inet_socket (const char *addr, const char *service, int flags)
{
  struct addrinfo hints;
  int fd;
  const char *node_addr;

  if (!service || !*service)
    RUNTIME_ERROR ("service is not specified");

  memset (&hints, 0, sizeof (struct addrinfo));
  hints.ai_family = AF_INET6; /* Prefer IPv6 address */
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = AI_PASSIVE;
  hints.ai_protocol = 0;

  node_addr = any (addr);

  if ((fd = _create_inet_socket (node_addr, service, flags, &hints)) < 0)
  {
    /* We are failed with IPv6, try IPv4 */
    hints.ai_family = AF_INET;
    fd = _create_inet_socket (node_addr, service, flags, &hints);
  }

  if (fd >= 0)
  {
      if (FLAG_ISSET(flags, C_SO_NOBLK))
      {
          if (set_nonblock (fd) == -1)
              EXIT_IF_FATAL (flags, "set_nonblock:%m");
      }

      if (FLAG_ISSET(flags, C_SO_SERVER))
      {
          if (listen (fd, NLISTEN) == -1)
              EXIT_IF_FATAL (flags, "listen:%m");
      }
  }
  else
  {
    EXIT_IF_FATAL (flags, "failed");
  }

  return fd;
}


void
unix_socket_set_ownership(const char *path, const char *user, const char *group)
{
    struct group *gr;
    struct passwd *pw;

    if (!user || !group || !path)
	return;

    ASSERT_STRERROR ((pw = getpwnam (user)) != NULL);
    ASSERT_STRERROR ((gr = getgrnam (group)) != NULL);
    ASSERT_STRERROR (chown(path, pw->pw_uid, gr->gr_gid) == 0);
    ASSERT_STRERROR (chmod(path,0660) == 0);
}


int
create_unix_socket (const char *path, int flags)
{
    int fd;

    if ((fd = socket(AF_UNIX,SOCK_STREAM,0)) < 0)
    {
        EXIT_IF_FATAL (flags, "socket:%m");
        return -1;
    }

    if (FLAG_ISSET (flags, C_SO_SERVER))
    {
        if (bind_unix(fd, path) < 0)
        {
            EXIT_IF_FATAL (flags, "can't bind");
            goto error;
        }

        if (listen (fd, NLISTEN) == -1)
        {
            EXIT_IF_FATAL (flags, "listen:%m");
            goto error;
        }
    }
    else
    {
        if (connect_unix(fd, path) < 0)
        {
            EXIT_IF_FATAL (flags, "connect_unix fail");
            goto error;
        }
    }

    if (FLAG_ISSET(flags, C_SO_NOBLK))
    {
        if(set_nonblock (fd) < 0)
            EXIT_IF_FATAL (flags,"set_nonblock:%m");
    }

    return fd;

error:
    close (fd);

    return -1;
}

void
close_socket (int fd)
{
  if (fd >= 0)
    close (fd);
}

int
socket_accept (int fd, char **remote_addr,int flags)
{
  int retcode;
  struct sockaddr_storage sa;
  socklen_t sa_len = sizeof (sa);
  int s;

  if ((fd == -1) || ((s = accept (fd, remote_addr?(struct sockaddr *)&sa:NULL, remote_addr?&sa_len:NULL)) == -1))
    return -1;

  close_on_exec(fd);

  if (remote_addr)
  {
    char hbuf[NI_MAXHOST];
    if ((retcode =
       getnameinfo ((struct sockaddr *)&sa, sa_len, hbuf, sizeof hbuf, NULL, 0,
		    NI_NUMERICHOST)) != 0)
    RUNTIME_ERROR ("getnameinfo:%s\n", gai_strerror (retcode));
    *remote_addr = g_strdup (hbuf);
  }

  if (FLAG_ISSET(flags, C_SO_NOBLK))
  {
    if (set_nonblock (s) == -1)
        EXIT_IF_FATAL (flags, "set_nonblock:%m");
  }

  return s;
}

int
socket_send (int fd, const void *buffer, size_t len)
{
  const char *p = (const char *)buffer;
  do
    {
      ssize_t count =
	TEMP_FAILURE_RETRY (send (fd, p, len, MSG_NOSIGNAL));

      if (count == -1)
	{
	  if (errno == EAGAIN)
	    continue;
	  else
	    return 1;
	}

      if (count == 0)
	return 1;
      else
	{
	  p += count;
	  len -= count;
	};
    }
  while (len);
  return 0;
}

int
socket_sendfile (int out_fd, int in_fd, size_t len)
{
  off_t offset = 0;

  do
    {
      ssize_t count =
	TEMP_FAILURE_RETRY (sendfile (out_fd, in_fd, &offset, len));

      if (count == -1)
	{
	  if (errno == EAGAIN)
	    continue;
	  else
	    return 1;
	}

      if (count == 0)
	return 1;
      else
	len -= count;
    }
  while (len);

  return 0;
}

int
socket_receive (int fd, buffer_t *msg_buf)
{
  char buf[BUFSIZ];
  int retcode = 0;

  ssize_t count = TEMP_FAILURE_RETRY (recv (fd, buf, sizeof (buf), 0));

  if (count == -1)
      retcode = (errno == EAGAIN) ? 0 : 1;
  else if (count == 0)
      retcode = 1;
  else
      g_string_append_len (msg_buf, buf, count);

  return retcode;
}

