#include <unistd.h>

#include <glib.h>

#include <vhttpd/tls.h>
#include <vhttpd/channel.h>
#include <vhttpd/tls_channel.h>

#include "private.h"
#include "tls_channel_private.h"

/*** functions */

static void
tls_channel_init (tls_channel_t *c, int fd, char *remote_addr,
                  tls_ctx_t * ctx, tls_session_t * session,
                  int flags)
{
    c->base.accept = FLAG_ISSET (flags, C_SO_SERVER) ? tls_channel_accept : NULL;
    c->base.send = tls_channel_send;
    c->base.sendfile = tls_channel_sendfile;
    c->base.receive = tls_channel_receive;
    c->base.destroy = destroy_tls_channel;
    c->base.fd = fd;
    c->base.remote_addr = remote_addr;
    c->base.type = CHANNEL_TYPE_TLS;
    c->base.flags = flags;
    c->ctx = ctx;
    c->session = session;
}


void
destroy_tls_channel (channel_t *c)
{
  tls_channel_t *tc = TLS_CHANNEL_T (c);

  if (tc->session)
    tls_session_delete (tc->session);
  if (tc->ctx)
    tls_ctx_delete (tc->ctx);

  close_socket (tc->base.fd);
  g_free (tc->base.remote_addr);

  g_free (tc);
}

tls_session_t *
tls_channel_session (channel_t *c)
{
  return c ? (TLS_CHANNEL_T (c))->session : NULL;
}

tls_ctx_t *
tls_channel_ctx (channel_t *c)
{
  return c ? (TLS_CHANNEL_T (c))->ctx : NULL;
}

int
tls_channel_send (channel_t *c, const char *buf, size_t len)
{
    return tls_session_send (tls_channel_session (c), buf, len);
}

int
tls_channel_sendfile (channel_t * c, int fd, size_t len)
{
    return tls_session_sendfile (tls_channel_session (c), fd, len);
}

int
tls_channel_receive (channel_t * c, buffer_t *buf)
{
    return tls_session_receive (tls_channel_session (c), buf);
}

channel_t *
tls_channel_accept (channel_t * c)
{
  char *remote_addr = NULL;
  tls_channel_t *ch_accepted;
  int fd;

  ch_accepted = g_new0 (tls_channel_t, 1);
  fd = socket_accept (channel_get_fd (c), &remote_addr,c->flags);
  tls_session_t *s = tls_session_new (fd, tls_channel_ctx (c));
  if (fd >= 0)
      tls_channel_init (ch_accepted, fd, remote_addr, NULL, s, 0);

  return CHANNEL_T(ch_accepted);
}

/* Note: server side supported only */
channel_t *
create_tls_channel_from_socket (int socket_fd, const char *addr, const char *keyfile,
				    const char *certfile, int flags)
{
  tls_channel_t *c;
  tls_ctx_t *ctx;

  if (socket_fd < 0)
      return NULL;

  ctx = tls_ctx_new (keyfile, certfile);
  c = g_new0 (tls_channel_t, 1);
  tls_channel_init (c, socket_fd, g_strdup (addr), ctx, NULL, flags);

  return CHANNEL_T (c);
}

/* Note: server side supported only */
channel_t *
create_tls_channel (const char *addr, const char *service, const char *keyfile,
		    const char *certfile, int flags)
{
  int s;

  if((s = create_inet_socket(addr, service, flags)) < 0)
      return NULL;

  return create_tls_channel_from_socket (s, addr, keyfile, certfile, flags);
}
